/*
** Copyright (C) 2003 Christophe Kalt
**
** This file is part of shmux
** see the LICENSE file for details on your rights.
**
** $Id$
*/

#if !defined(_SIGLIST_H_)
# define _SIGLIST_H_

int getsignumbyname(char *);

#endif
