/*
** Copyright (C) 2002, 2003, 2004, 2005, 2006 Christophe Kalt
**
** This file is part of shmux
** see the LICENSE file for details on your rights.
**
** $Id$
*/

#if !defined(_UNITS_H_)
# define _UNITS_H_

u_int unit_time(char *);
char *unit_rtime(u_int);

#endif
