/*
** Copyright (C) 2003 Christophe Kalt
**
** This file is part of shmux,
** see the LICENSE file for details on your rights.
**
** This file was automatically generated by siglist.awk using awk.
**
** $Id: signals.awk,v 1.1 2003-04-26 01:39:51 kalt Exp $
*/

struct signame {
    char *name;
    int num;
};

struct signame signame[] = {
    { NULL,	 0 }
};
